!
! exportbands --- read in a band structure from CRYSTAL2003 and write out
!       a file suitable for cabplot
! Christian Buth
! 12.11.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
!
!
! Version 1.00: 16.11.2004 (Write bands from CRYSTAL)
! Version 1.10: 22.02.2005 (Read data from CRYSTAL's F25 unit,
!       handling of multiple paths)
!


      program exportbands
      use fortrantl
      use physconst
      use bandstructure
      implicit none

      type tcrystalbandfile
        ! Additionally required or calculated
        integer iss,       &            ! Global shrinking factor (denominator
                                        ! for k-point specification). Not given
                                        ! in crystal band structure files!
     &          numkpoints              ! Number of k-points along the path.
                                        ! If a preceeding path shares the end point
                                        ! with this path, the shared point is counted.

        ! First record of a path of a band structure
        character(3) signature          ! '-%-'
        integer ihferm                  ! Type of calculation:
                                        ! 0: closed-shell, insulating system
                                        ! 1: open-shell, conducting system
                                        ! 2: closed-shell, insulating system
                                        ! 3: open-shell, conducting system
        character(4) datatype           ! 'BAND', 'DOSS, 'EMDL', 'EMDP'
        integer numBands, &             ! Number of bands
     &          numkdata                ! Number of k-points along the path written
                                        ! by CRYSTAL to the file. If several
                                        ! calls to BAND in CRYSTAL are made to compute
                                        ! individual paths, the shared end points
                                        ! between paths are replicated!
                                        ! If a contiguous path is computed with a
                                        ! a single call to BAND in CRYSTAL,
                                        ! preceeding path shares the end point
                                        ! with the starting point of the current
                                        ! path, the point is not counted in the
                                        ! current path.
        real(dblprec) unused, &         ! Unused for band structures
     &          distkspace,   &         ! Distance between two adjacent h-points
                                        ! in k-space
     &          fermienergy             ! The Fermi energy

        ! Second record of a path of a band structure
        real(dblprec) minEnergy, &      ! Minimum energy in the path
     &          maxEnergy               ! Maximum energy in the path

        ! Third record of a path of a band structure
        type(tlatticeindex) begk, endk  ! First and last point of the path.
                                        ! Coordinates with respect to
                                        ! the reciprocal lattice vectors of
                                        ! the primitive unit cell

        ! Fourth record the band structure data
        real(dblprec), pointer :: bands(:,:)
      end type tcrystalbandfile


      logical individualband

      character(1024), parameter :: fileCrystal = 'c.LiF.11.f25', &
     &                              fileWannier = 'unsinn.dat'
      character(3), parameter :: cryf25sign = '-%-'
      character(4), parameter :: crybandsign = 'BAND'

      integer, parameter :: fcrystal = 50, &
     &                      fwannier = 51
      integer, parameter :: maxCryPaths = 20    ! CRYSTAL handles at most that many
                                                ! paths
      integer currPath, currkpoint, numBrillouinPaths, numOccupied, &
     &        periodicity, spacegroup, shiftfirst, i, j

!      real(dblprec) shiftBands

      type(tlatticevector) kpoint, lbeg, lend, linc
!      type(tlatticetraf) mt
      type(tfbands) fbands
      type(tbandstructure) crystalbands
      type(tcrystalbandfile) crybandf(maxCryPaths)


      write (*,*) 'Export bands from CRYSTAL to WANNIER'
      write (*,*) '(c) 2005 Christian Buth'
      write (*,*)

      ! Read band structure from Crystal
      write (*,*) 'Reading Crystal bands...'
      write (*,*) 'CRYSTAL bands are in ', trim(fileCrystal)
      write (*,*)


        individualband = .true.
        periodicity = 3
        spacegroup = 225


      open(unit=fcrystal, file=fileCrystal, status='old', action='read')

      currPath = 1

      do

      crybandf%iss = 8

      ! First record of path through Brillouin zone
      ! Format: '(A3,I1,A4,2I4,1P,3E12.5)'
      read(fcrystal,'(A3,I1,A4,2I4,1P,3E12.5)', iostat=i) crybandf(currPath)%signature, &
     &          crybandf(currPath)%ihferm, crybandf(currPath)%datatype, &
     &          crybandf(currPath)%numBands, &
     &          crybandf(currPath)%numkdata, crybandf(currPath)%unused, &
     &          crybandf(currPath)%distkspace, crybandf(currPath)%fermienergy
      ! End of file?
      if (i < 0) then
        exit
      end if
      crybandf(currPath)%fermienergy = crybandf(currPath)%fermienergy * HartreeToeV
      write (*,*) crybandf(currPath)%signature, crybandf(currPath)%ihferm, &
     &          crybandf(currPath)%datatype, &
     &          crybandf(currPath)%numBands, &
     &          crybandf(currPath)%numkdata, crybandf(currPath)%unused, &
     &          crybandf(currPath)%distkspace, crybandf(currPath)%fermienergy
      if (crybandf(currPath)%signature /= cryf25sign) then
        stop 'Wrong signature: this is not a CRYSTAL F25 file'
      else if (crybandf(currPath)%ihferm /= 0) then
        stop 'The band structure is not from a closed-shell, insulating system!'
      else if (crybandf(currPath)%datatype /= crybandsign) then
        stop 'Wrong content: this is not a CRYSTAL band structure file'
      end if

      ! Second record of path through Brillouin zone
      ! Format: '(1P,6E12.5,)'
      read(fcrystal,'(1P,2E12.5)') crybandf(currPath)%minEnergy, &
     &          crybandf(currPath)%maxEnergy
      crybandf(currPath)%minEnergy = crybandf(currPath)%minEnergy * HartreeToeV
      crybandf(currPath)%maxEnergy = crybandf(currPath)%maxEnergy * HartreeToeV
      write (*,*) crybandf(currPath)%minEnergy, crybandf(currPath)%maxEnergy

      ! Third record of path through Brillouin zone
      ! Format: '(6I3,1X,A36)'
      read(fcrystal,'(6I3)') crybandf(currPath)%begk%i, crybandf(currPath)%endk%i
      write (*,*) crybandf(currPath)%begk, crybandf(currPath)%endk

      ! Read bands for the given path through the Brillouin zone
      ! Format: '(1P,6E12.5)'
      allocate(crybandf(currPath)%bands(crybandf(currPath)%numBands, &
     &         crybandf(currPath)%numkdata))
      read(fcrystal,'(1P,6E12.5)') ((crybandf(currPath)%bands(i,j), &
     &          i=1, crybandf(currPath)%numBands), j=1, crybandf(currPath)%numkdata)
      crybandf(currPath)%bands(:,:) = crybandf(currPath)%bands(:,:) * HartreeToeV

      ! Calculate number of k-points, counting shared points
      ! between paths appropriately.
      crybandf(currPath)%numkpoints = crybandf(currPath)%numkdata
      if (.not. individualband .and. currPath > 1) then
        ! Does the current path begin where the previous one ended?
        if (crybandf(currPath - 1)%endk%i(1) == crybandf(currPath)%begk%i(1) .and. &
     &      crybandf(currPath - 1)%endk%i(2) == crybandf(currPath)%begk%i(2) .and. &
     &      crybandf(currPath - 1)%endk%i(3) == crybandf(currPath)%begk%i(3)) then
          ! The contiguous path calculated by a single BAND call in CRYSTAL
          ! and not by individual calls to BAND!
          crybandf(currPath)%numkpoints = crybandf(currPath)%numkpoints + 1
        end if
      end if
      write (*,*) 'Number of kpoints ', crybandf(currPath)%numkdata, &
     &          crybandf(currPath)%numkpoints


      currPath = currPath + 1
      if (currPath > maxCryPaths) then
        stop 'CRYSTAL handles no more than maxCryPaths but there are more in the file!'
      end if

      end do


      numBrillouinPaths = currPath - 1
      write (*,*) 'Number of paths ', numBrillouinPaths

      close(fcrystal)

      ! Determine the number of occupied bands (works only for insulators!)
      numOccupied = 0
      do i=1, crybandf(1)%numBands
        if (crybandf(1)%bands(i,1) < crybandf(1)%fermienergy) then
          numOccupied = numOccupied + 1
        end if
      end do
      write (*,*) 'Number of occupied bands ', numOccupied

      ! Get paths though the Brillouin zone
      allocate(fbands%paths(numBrillouinPaths))
      fbands%fock = .true.
      fbands%numPaths = numBrillouinPaths
      fbands%startOccupied = 1
      fbands%numOccupied = numOccupied
      fbands%numOrbitals = crybandf(1)%numBands
      fbands%periodicity = periodicity
      fbands%spacegroup  = 0

!      mt = getRecSpTrafPrimCell(periodicity, spacegroup)
      do currPath=1, numBrillouinPaths
        lbeg%v = real(crybandf(currPath)%begk%i, dblprec) &
     &           / real(crybandf(currPath)%iss, dblprec)
!        lbeg%v = matmul(mt%m, lbeg%v)   ! Transformation to conventional unit cell
        lend%v = real(crybandf(currPath)%endk%i, dblprec) &
     &           / real(crybandf(currPath)%iss, dblprec)
!        lend%v = matmul(mt%m, lend%v)   ! Transformation to conventional unit cell
        linc%v = (lend%v - lbeg%v) / real(crybandf(currPath)%numkpoints - 1, dblprec)
        call initbrillouinpathspec(fbands%paths(currPath), periodicity, &
     &          spacegroup, lbeg, lend, linc)
      end do
      call reportbrillouinpathspec(fbands%paths, numBrillouinPaths, periodicity)


/******************************************************
      ! Read band structure from Crystal
      call initbandstructure(crystalbands, fbands)
      open(unit=fcrystal, file=fileCrystal, status='old', action='read')

      do currPath=1, numBrillouinPaths
        kpoint = fbands%paths(currPath)%beg
        allocate(crystalbands%paths(currPath)%bands( &
     &           crystalbands%paths(currPath)%spec%numkpoints) )

        do currkpoint=1, crystalbands%paths(currPath)%spec%numkpoints
          allocate(crystalbands%paths(currPath)%bands(currkpoint)%energies(crybandf(1)%numBands))
          nullify (crystalbands%paths(currPath)%bands(currkpoint)%polestrengths)
          crystalbands%paths(currPath)%bands(currkpoint)%kpoint = kpoint

!          write (*,*) 'Reading kpoint ', currkpoint
!          read (fcrystal,'(10000(F13.5))') dummykpoint%a,                          &
!     &        (crystalbands%paths(currPath)%bands(currkpoint)%energies(i), i=1, crybandf(1)%numBands)
          read (fcrystal,*) dummykpoint%v(1),                          &
     &        (crystalbands%paths(currPath)%bands(currkpoint)%energies(i), i=1, crybandf(1)%numBands)
!          write (*,*) dummykpoint%v(1),                          &
!     &        (crystalbands%paths(currPath)%bands(currkpoint)%energies(i), i=1, crybandf(1)%numBands)
!                stop
!          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) = &
!     &          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) * eVToHartree

!          kpoint%a = kpoint%a + crystalbands%paths(currPath)%spec%inc%a
!          kpoint%b = kpoint%b + crystalbands%paths(currPath)%spec%inc%b
!          kpoint%c = kpoint%c + crystalbands%paths(currPath)%spec%inc%c
          kpoint%v = kpoint%v + crystalbands%paths(currPath)%spec%inc%v
        end do
      end do
      close(fcrystal)
******************************************************/


      ! Convert band structure from Crystal
      call initbandstructure(crystalbands, fbands)

      ! Traverse all paths
      do currPath=1, numBrillouinPaths
        kpoint = fbands%paths(currPath)%beg
        allocate(crystalbands%paths(currPath)%bands &
     &          (crystalbands%paths(currPath)%spec%numkpoints))

        ! Need a special treatment of first point in a path which
        ! has to be replicated, if it is the end point of the
        ! previous path.
        if (crybandf(currPath)%numkpoints == crybandf(currPath)%numkdata + 1) then
          allocate(crystalbands%paths(currPath)%bands(1)% &
     &             energies(crybandf(currPath)%numBands))
          nullify(crystalbands%paths(currPath)%bands(1)%polestrengths)
          crystalbands%paths(currPath)%bands(1)%kpoint = kpoint
          crystalbands%paths(currPath)%bands(1)%energies(:) = &
      &         crystalbands%paths(currPath - 1)%bands(crystalbands%paths(currPath &
      &         - 1)%spec%numkpoints)%energies(:)
          crystalbands%paths(currPath)%bands(1)%numBands = &
     &          crybandf(currPath)%numBands
          kpoint%v = kpoint%v + crystalbands%paths(currPath)%spec%inc%v
          shiftfirst = 1
        else
          shiftfirst = 0
        end if

        ! Traverse all k-points in a path
        do currkpoint = 1 + shiftfirst, crystalbands%paths(currPath)%spec%numkpoints
          allocate(crystalbands%paths(currPath)%bands(currkpoint)% &
     &             energies(crybandf(currPath)%numBands))
          nullify(crystalbands%paths(currPath)%bands(currkpoint)%polestrengths)
          crystalbands%paths(currPath)%bands(currkpoint)%kpoint = kpoint
          crystalbands%paths(currPath)%bands(currkpoint)%numBands = &
     &          crybandf(currPath)%numBands

          ! Copy all energies for a k-point
          do i=1, crybandf(currPath)%numBands
            crystalbands%paths(currPath)%bands(currkpoint)%energies(i) = &
      &         crybandf(currPath)%bands(i, currkpoint - shiftfirst)
          end do

!          read (fcrystal,'(10000(F13.5))') dummykpoint%a,                          &
!     &        (crystalbands%paths(currPath)%bands(currkpoint)%energies(i), i=1, crybandf(1)%numBands)
!          read (fcrystal,*) dummykpoint%v(1),                          &
!     &        (crystalbands%paths(currPath)%bands(currkpoint)%energies(i), i=1, crybandf(1)%numBands)
!          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) = &
!     &          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) * eVToHartree

          kpoint%v = kpoint%v + crystalbands%paths(currPath)%spec%inc%v
        end do
      end do


      ! Write band structure for Wannier
      write (*,*) 'Writing Wannier bands...'
      call openbands(fbands, fwannier, fileWannier, 'write')

      ! Traverse all paths
      do currPath=1, numBrillouinPaths
        call putbrillouinpath(fbands, currPath)
        kpoint = fbands%paths(currPath)%beg

        ! Traverse all k-points in a path
        do currkpoint=1, fbands%paths(currPath)%numkpoints
!          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) = &
!     &          crystalbands%paths(currPath)%bands(currkpoint)%energies(:) * HartreetoeV &
!     &          + shiftBands
          call putbandrecord(fbands, kpoint, crystalbands%paths(currPath)% &
     &          bands(currkpoint)%numBands, &
     &          crystalbands%paths(currPath)%bands(currkpoint)%energies)

          kpoint%v = kpoint%v + fbands%paths(currPath)%inc%v
        end do
      end do

      call closebands(fbands)
      write (*,*) 'Wannier bands are in ',trim(fileWannier)
      write (*,*)

      ! Clean up
      !call exitbandstructure(crystalbands)
      do i=1, numBrillouinPaths
        deallocate(crybandf(i)%bands)
      end do
      end program exportbands
