!
! bandsfile --- Reading and writing Hartree-Fock or CO-ADC band structures
! Christian Buth
! 04.07.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
! Version 1.00: 04.07.2004 (Reading and writing Hartree-Fock or CO-ADC band structures)
! Version 1.10: 25.08.2004 (Space groups)
!


!
! Includes
!

#include "assert.h"


!
! Module
!

      module bandsfile
      use brillouinpathspec
      use wannier
      implicit none


!
! Types
!

      type tfbands
        logical action,        &           ! .true. for reading, .false. for writing
     &          fock                       ! .true. for Hartree-Fock band structure, .false. for CO-ADC band structure
        integer unit,          &           ! FORTRAN file unit
     &          startOccupied, &           ! First occupied orbital given
     &          numOrbitals,   &           ! Total number of orbitals
     &          numOccupied,   &           ! Total number of occupied orbitals
     &          periodicity,   &           ! Periodicity of the system
     &          spacegroup,    &           ! Spacegroup of the system
     &          numPaths                   ! Number of paths through the Briullouin zone

      type(tbrillouinpathspec), pointer :: paths(:)
      end type tfbands


!
! Module procedures
!

      contains


!
! openbands --- Open band structure files
!

      subroutine openbands(fbands, unit, file, action)
      implicit none

      character(*), intent(in) :: file, action
      integer, intent(in) :: unit
      type(tfbands), intent(inout) :: fbands

      character(7) status
      character(11) signature
      integer currPath

      ! Initialize openbands
      fbands%unit = unit
      select case (trim(action))
        case ('read')
          fbands%action = .true.
          status='old'
        case ('write')
          fbands%action = .false.
          status='replace'
        case default
          write (*,*) 'openbands: files can only be opened for reading or writing!'
          stop 'Error in openbands!'
      end select
      open(unit=unit, file=trim(file), status=trim(status), action=trim(action))

      ! Opened for reading
      if (fbands%action) then

        ! Read file header
        read (unit,*) signature
        if (trim(signature) /= 'Bands_V1.00') then
          write (*,*) 'openbands: the file ', trim(file), 'is not a band structure file V1.00!'
          stop 'Error in openbands!'
        end if
        read (unit,*) fbands%fock, fbands%startOccupied,      &
     &                fbands%numOccupied,  &
     &                fbands%numOrbitals, fbands%periodicity, &
     &                fbands%spacegroup, fbands%numPaths
        write (*,*) 'fbands%fock, fbands%startOccupied, fbands%numOccupied, fbands%numOrbitals,', &
     &               'fbands%periodicity, fbands%spacegroup, fbands%numPaths'
        write (*,*) fbands%fock, fbands%startOccupied, fbands%numOccupied, fbands%numOrbitals, &
     &              fbands%periodicity, fbands%spacegroup, fbands%numPaths

        ! Get paths though the Brillouin zone
        allocate(fbands%paths(fbands%numPaths))
        do currPath=1, fbands%numPaths
          call readbrillouinpathspec(fbands%paths(currPath), unit, &
     &                      fbands%periodicity, fbands%spacegroup)
        end do

      ! Opened for writing
      else

        write (unit,*) 'Bands_V1.00'
        write (unit,*) fbands%fock, fbands%startOccupied,      &
     &                 fbands%numOccupied,  &
     &                 fbands%numOrbitals, fbands%periodicity, &
     &                 fbands%spacegroup, fbands%numPaths
        do currPath=1, fbands%numPaths
          call writebrillouinpathspec(fbands%paths(currPath), unit, fbands%periodicity)
        end do

      end if
      end subroutine openbands



!
! closebands --- Close band structure files
!

      subroutine closebands(fbands)
      implicit none

      type(tfbands), intent(inout) :: fbands

      close(unit=fbands%unit)
      deallocate(fbands%paths)
      end subroutine closebands


!
! putbrillouinpath --- Start writing a new path through the Brillouin zone in a band structure file
!

      subroutine putbrillouinpath(fbands, currPath)
      implicit none

      integer, intent(in) :: currPath
      type(tfbands), intent(inout) :: fbands

      assert(.not. fbands%action)
      write (fbands%unit,*) currPath, fbands%paths(currPath)%numkpoints
      end subroutine putbrillouinpath


!
! putkpointheader --- Start writing a new record for a kpoint of a path through the Brillouin zone
!       in a band structure file
!

      subroutine putkpointheader(fbands, kpoint, numentries)
      implicit none

      integer, intent(in), optional :: numentries
      type(tlatticevector), intent(in) :: kpoint
      type(tfbands), intent(inout) :: fbands

      assert(.not. fbands%action)
      if (fbands%fock) then
        write (fbands%unit,*) kpoint
      else
        write (fbands%unit,*) kpoint, numentries
      end if
      end subroutine putkpointheader


!
! putbandpoint --- Put the energy and pole strength of a single band for a single kpoint in
!       a path through the Brillouin zone to a band structure file
!

      subroutine putbandpoint(fbands, energy, polestrength)
      implicit none

      real(dblprec), intent(in) :: energy
      real(dblprec), intent(in), optional :: polestrength
      type(tfbands), intent(inout) :: fbands

      assert(.not. fbands%action)
      if (fbands%fock) then
        write (fbands%unit,*) energy
      else
        write (fbands%unit,*) energy, polestrength
      end if
      end subroutine putbandpoint


!
! putbandrecord --- Put energies and pole strengths of for complete band structure for a
!       single kpoint in a path through the Brillouin zone to a band structure file
!

      subroutine putbandrecord(fbands, kpoint, numentries, energies, polestrengths)
      implicit none

      integer, intent(in) :: numentries
      real(dblprec), intent(in) :: energies(:)
      real(dblprec), intent(in), optional :: polestrengths(:)
      type(tlatticevector), intent(in) :: kpoint
      type(tfbands), intent(inout) :: fbands

      integer i

      assert(.not. fbands%action)
      call putkpointheader(fbands, kpoint, numentries)
      if (fbands%fock) then
        do i=1, numentries
          call putbandpoint(fbands, energies(i))
        end do
      else
        do i=1, numentries
          call putbandpoint(fbands, energies(i), polestrengths(i))
        end do
      end if
      end subroutine putbandrecord


!
! getbrillouinpath --- Start reading a new path through the Brillouin zone in a band structure file
!

      subroutine getbrillouinpath(fbands, currPath)
      implicit none

      integer, intent(in) :: currPath
      type(tfbands), intent(inout) :: fbands

      integer pathno, kpointsno

      assert(fbands%action)
      read (fbands%unit,*) pathno, kpointsno
      if (pathno /= currPath) then
        write (*,*) 'getbrillouinpath: Maleformed band structure V1.00 file: next path is ', &
     &              pathno, ' but expect path No ', currPath, '!'
        stop 'Error in getbrillouinpath!'
      end if
      if (kpointsno /= fbands%paths(currPath)%numkpoints) then
        write (*,*) 'getbrillouinpath: Maleformed band structure V1.00 file: path contains ', &
     &              kpointsno, ' kpoints where ', fbands%paths(currPath)%numkpoints, ' are expected!'
        stop 'Error in getbrillouinpath!'
      end if
      end subroutine getbrillouinpath


!
! getkpointheader --- Start reading a new record for a kpoint of a path through the Brillouin zone
!       in a band structure file
!

      subroutine getkpointheader(fbands, kpoint, numentries)
      implicit none

      integer, intent(out), optional :: numentries
      type(tlatticevector), intent(out) :: kpoint
      type(tfbands), intent(inout) :: fbands

      assert(fbands%action)
      if (fbands%fock) then
        read (fbands%unit,*) kpoint
        if (present(numentries)) then
          numentries = fbands%numOrbitals
        end if
      else
        assert(present(numentries))
        read (fbands%unit,*) kpoint, numentries
      end if
      end subroutine getkpointheader


!
! getbandpoint --- Get the energy and pole strength of a single band for a single kpoint in
!       a path through the Brillouin zone from a band structure file
!

      subroutine getbandpoint(fbands, energy, polestrength)
      implicit none

      real(dblprec), intent(out) :: energy
      real(dblprec), intent(out), optional :: polestrength
      type(tfbands), intent(inout) :: fbands

      real(dblprec) ps

      assert(fbands%action)
      if (fbands%fock) then
        read (fbands%unit,*) energy
        ps = 1
      else
        read (fbands%unit,*) energy, ps
      end if
      if (present(polestrength)) then
        polestrength = ps
      end if
      end subroutine getbandpoint


!
! getbandrecord --- Get energies and pole strengths of a complete band structure for a
!       single kpoint in a path through the Brillouin zone from a band structure file
!

      subroutine getbandrecord(fbands, kpoint, energies, polestrengths, numentries)
      implicit none

      integer, intent(out), optional :: numentries
      real(dblprec), intent(out) :: energies(:)
      real(dblprec), intent(out), optional :: polestrengths(:)
      type(tlatticevector), intent(out) :: kpoint
      type(tfbands), intent(inout) :: fbands

      integer i, entriesno

      assert(fbands%action)
      call getkpointheader(fbands, kpoint, entriesno)
      if (present(numentries)) then
        numentries = entriesno
      end if
      do i=1, entriesno
        if (present(polestrengths)) then
          call getbandpoint(fbands, energies(i), polestrengths(i))
        else
          call getbandpoint(fbands, energies(i))
        end if
      end do
      end subroutine getbandrecord
      end module bandsfile
