!
! bandstructure --- Manipulate band structures
! Christian Buth
! 27.02.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
! Version 1.00: 27.02.2004 (Read in the BANDS???.DAT file)
! Version 1.10: 07.03.2004 (Added the selection of a subset of bands. Now all
!       data in BANDS???.DAT files are interpreted correctly)
! Version 1.20: 28.06.2004 (Added properties reading only)
! Version 1.30: 03.07.2004 (Support for three dimensional crystals)
! Version 1.40: 12.07.2004 (Module created for the tbandstructure type based on
!       wannierbands [removed only properties reading beforehand] subroutine to
!       read Wannier's Hartree-Fock band structure, added reading of band
!       structures from files created with the bandsfile module, added
!       reportbrillouinspecs and fermienergy)
! Version 1.50: 25.08.2004 (Spacegroups, metric and coordinate length)
! Version 1.60: 22.09.2004 (Calculation of band widths for Hartree-Fock bands)
! Version 1.70: 12.11.2004 (Initialize band structure file from a tfbands type)
! Version 1.80: 24.02.2005 (Search k-points in Brillouin paths)
!


      module bandstructure
      use bandsfile
      implicit none


!
! Types
!

      type tbspoint
        integer pathno, kpointno, bandno
      end type tbspoint


      type tbandsatkpoint
        integer numBands
        type(tlatticevector) kpoint
        real(dblprec), pointer :: energies(:),        &
     &                            polestrengths(:)
      end type tbandsatkpoint


      type tbrillouinpath
        type(tbrillouinpathspec) spec
        type(tbandsatkpoint), pointer :: bands(:)
      end type tbrillouinpath


      type tbandstructure
        logical fock
        integer numPaths,        &
     &          totalnumkpoints, &
     &          startOccupied,   &
     &          numOccupied,     &
     &          numOrbitals,     &
     &          spacegroup,      &
     &          periodicity
        real(dblprec) metriclength, coordlength             ! Length of all paths
        type(tbrillouinpath), pointer :: paths(:)
      end type tbandstructure


!
! Interface
!

      interface initbandstructure
        module procedure initbandstructure_tfbands
        module procedure initbandstructure_bandsfile
        module procedure initbandstructure_wannier
      end interface initbandstructure

      interface reportbrillouinpathspec
        module procedure reportbrillouinpathspec_array
        module procedure reportbrillouinpathspec_bs
      end interface reportbrillouinpathspec


!
! Module procedures
!

      contains

!
! initbandstructure_tfbands --- initialize band structure file from a tfbands type
!

      subroutine initbandstructure_tfbands(bs, fbands)
      implicit none

      type(tfbands), intent(in) :: fbands
      type(tbandstructure), intent(out) :: bs
      integer currPath

      ! Initialize band structure
      bs%fock            = fbands%fock
      bs%numPaths        = fbands%numPaths
      bs%totalnumkpoints = 0
      bs%numOccupied     = fbands%numOccupied
      bs%numOrbitals     = fbands%numOrbitals
      bs%startOccupied   = fbands%startOccupied
      bs%periodicity     = fbands%periodicity
      bs%spacegroup      = fbands%spacegroup
      bs%metriclength    = 0
      bs%coordlength     = 0
      allocate(bs%paths(bs%numPaths))

      do currPath=1, fbands%numPaths
        bs%paths(currPath)%spec = fbands%paths(currPath)
        bs%metriclength = bs%metriclength + bs%paths(currPath)%spec%metriclength
        bs%coordlength  = bs%coordlength  + bs%paths(currPath)%spec%coordlength
        bs%totalnumkpoints = bs%totalnumkpoints + bs%paths(currPath)%spec%numkpoints
        write (*,*) 'bs%metriclength, bs%coordlength'
        write (*,*) bs%metriclength, bs%coordlength
        nullify(bs%paths(currPath)%bands)
      end do

      write (*,'(A,I4)') 'Total number of kpoints along the paths ', bs%totalnumkpoints
      end subroutine initbandstructure_tfbands


!
! initbandstructure_bandsfile --- read a band structure file (module bandsfile)
!       into memory
!

      subroutine initbandstructure_bandsfile(bs, unit, file)
      implicit none

      character(*), intent(in) :: file
      integer, intent(in) :: unit
      type(tbandstructure), intent(out) :: bs
      type(tfbands) fbands

      integer currkpoint, currBand, currPath, numBands
      type(tlatticevector) kpoint

      ! Initialize band structure
      call openbands(fbands, unit, file, 'read')
      call initbandstructure_tfbands(bs, fbands)

      do currPath=1, fbands%numPaths
        call getbrillouinpath(fbands, currPath)
        allocate(bs%paths(currPath)%bands(bs%paths(currPath)%spec%numkpoints))

        do currkpoint=1, bs%paths(currPath)%spec%numkpoints
          call getkpointheader(fbands, kpoint, numBands)
          bs%paths(currPath)%bands(currkpoint)%kpoint   = kpoint
          bs%paths(currPath)%bands(currkpoint)%numBands = numBands
          allocate(bs%paths(currPath)%bands(currkpoint)%energies(numBands))

          ! If the file contains a Hartree-Fock band structure, all pole strengths
          ! are unity and are not stored
          if (bs%fock) then

            nullify(bs%paths(currPath)%bands(currkpoint)%polestrengths)
            do currBand=1, numBands
              call getbandpoint(fbands, &
     &          bs%paths(currPath)%bands(currkpoint)%energies(currBand))
            end do

          else

            allocate(bs%paths(currPath)%bands(currkpoint)%polestrengths(numBands))
            do currBand=1, numBands
              call getbandpoint(fbands, &
     &          bs%paths(currPath)%bands(currkpoint)%energies(currBand), &
     &          bs%paths(currPath)%bands(currkpoint)%polestrengths(currBand))
            end do

          end if
        end do

!        write (*,'(A,I2,A,I3)') 'Read path No ', currPath, ':', &
!     &          bs%paths(currPath)%spec%numkpoints
      end do

      call closebands(fbands)
      end subroutine initbandstructure_bandsfile


!
! initbandstructure_wannier --- read a band structure file from the Wannier program into memory
!

      subroutine initbandstructure_wannier(bs, unit, files,  &
     &          numBrillouinPaths, numOccupied, &
     &          periodicity, bpaths)
      implicit none

      character(*), intent(in) :: files(:)
      integer, intent(in) :: unit, numBrillouinPaths, numOccupied, &
     &          periodicity
      type(tbandstructure), intent(out) :: bs
      type(tbrillouinpathspec), intent(in) :: bpaths(:)
      integer currPath

      ! Initialize band structure
      bs%fock            = .true.
      bs%numPaths        = numBrillouinPaths
      bs%totalnumkpoints = 0
      bs%periodicity     = periodicity
      bs%spacegroup      = 0
      bs%metriclength    = 0
      bs%coordlength     = 0
      allocate(bs%paths(bs%numPaths))

      do currPath=1, numBrillouinPaths

        bs%paths(currPath)%spec = bpaths(currPath)
        bs%metriclength = bs%metriclength + bs%paths(currPath)%spec%metriclength
        bs%coordlength  = bs%coordlength  + bs%paths(currPath)%spec%coordlength
        call getwannierbands(bs%paths(currPath), periodicity, unit, &
     &          files(currPath), bs%startOccupied, bs%numOrbitals)
        bs%numOccupied = numOccupied - bs%startOccupied - 1
!       write (*,'(A,I2,A,I3)') 'Read path No ', currPath, ':', &
!     &          bs%paths(currPath)%spec%numkpoints
        bs%totalnumkpoints = bs%totalnumkpoints + bs%paths(currPath)%spec%numkpoints

      end do

      write (*,'(A,I4)') 'Total number of kpoints along the paths ', bs%totalnumkpoints
      end subroutine initbandstructure_wannier


!
! getwannierbands --- Read a BANDS???.DAT file
!

      subroutine getwannierbands(bpath, periodicity, unit, file, &
     &          startOccupied, numOrbitals)
      implicit none

      character(len=*), intent(in) :: file
      integer, intent(in) :: periodicity, unit
      integer, intent(out) :: startOccupied, numOrbitals
      type(tbrillouinpath), intent(inout) :: bpath

      character(132) inpLine
      integer i, currkpoint, currBand, endVirtual, dummy, bandNo, &
     &        numkpoints, recordNo
      real(dblprec) dummy1, dummy2, dummy3, dummy4

      ! Open the Hartree-Fock band structure produced by the Wannier program
      open (unit=unit, file=file, form='formatted', action='read')
      read (unit,*) startOccupied, endVirtual, dummy
      write (*,*) 'startWannierOccupied, endWannierVirtual, dummy'
      write (*,*) startOccupied, endVirtual, dummy
      numOrbitals = endVirtual - startOccupied + 1

      ! Skip empty lines
      recordNo = 2
      do
        read (unit,*) inpLine
        if (verify(inpLine,' ') /= 0) then
          backspace(unit)
          exit
        end if
        recordNo = recordNo + 1
      end do

      ! Read band header
      read (unit,'(A11,I)') inpLine, bandNo

      ! Count the number of k points for a band
      numkpoints = 0
      do
        read (unit,'(A)') inpLine
        if (verify(inpLine,' ') == 0) then
          rewind(unit)
          do i=1, recordNo
            read (unit,*)
          end do
          exit

        else

          backspace(unit)
          select case (periodicity)
            case (0)
            case (1)
              read (unit,*) dummy1, dummy2
            case (2)
              read (unit,*) dummy1, dummy2, dummy3
            case (3)
              read (unit,*) dummy1, dummy2, dummy3, dummy4
          end select
        end if
        numkpoints = numkpoints + 1
      end do

      write (*,*) 'recordNo, numWannierkpoints'
      write (*,*) recordNo, bpath%spec%numkpoints
      if (bpath%spec%numkpoints /= numkpoints) then
        write (*,*) 'bpath%spec%numkpoints /= numkpoints'
        stop 'bpath%spec%numkpoints /= numkpoints'
      end if

      ! Allocate band structure memory
      allocate(bpath%bands(bpath%spec%numkpoints))
      do currkpoint=1, bpath%spec%numkpoints
        bpath%bands(currkpoint)%numBands = numOrbitals
        allocate(bpath%bands(currkpoint)%energies(numOrbitals))
        nullify(bpath%bands(currkpoint)%polestrengths)
      end do

      ! Read in all bands
      do currBand=1, numOrbitals
        ! Skip empty lines
        do
          read (unit,*) inpLine
          if (verify(inpLine,' ') /= 0) then
            backspace(unit)
            exit
          end if
        end do

        read (unit,'(A11,I)') inpLine, bandNo
        if (bandNo /= currBand + startOccupied - 1) then
          write (*,*) bandNo, currBand + startOccupied - 1, inpLine
          stop 'bandNo /= currBand + startOccupied - 1'
        end if

        do currkpoint=1, bpath%spec%numkpoints
          ! Wannier stores the band energies in electronvolt
          select case (periodicity)
            case (0)
            case (1)
!              read (unit,*) bpath%bands(currkpoint)%kpoint%a, &
!     &             bpath%bands(currkpoint)%energies(currBand)
!              bpath%bands(currkpoint)%kpoint%b = 0
!              bpath%bands(currkpoint)%kpoint%c = 0
              read (unit,*) bpath%bands(currkpoint)%kpoint%v(1), &
     &             bpath%bands(currkpoint)%energies(currBand)
              bpath%bands(currkpoint)%kpoint%v(2) = 0
              bpath%bands(currkpoint)%kpoint%v(3) = 0
            case (2)
!              read (unit,*) bpath%bands(currkpoint)%kpoint%a, &
!     &                      bpath%bands(currkpoint)%kpoint%b, &
!     &             bpath%bands(currkpoint)%energies(currBand)
!              bpath%bands(currkpoint)%kpoint%c = 0
              read (unit,*) bpath%bands(currkpoint)%kpoint%v(1), &
     &                      bpath%bands(currkpoint)%kpoint%v(2), &
     &             bpath%bands(currkpoint)%energies(currBand)
              bpath%bands(currkpoint)%kpoint%v(3) = 0
            case (3)
!              read (unit,*) bpath%bands(currkpoint)%kpoint%a, &
!     &                      bpath%bands(currkpoint)%kpoint%b, &
!     &                      bpath%bands(currkpoint)%kpoint%c, &
!     &             bpath%bands(currkpoint)%energies(currBand)
              read (unit,*) bpath%bands(currkpoint)%kpoint%v(1), &
     &                      bpath%bands(currkpoint)%kpoint%v(2), &
     &                      bpath%bands(currkpoint)%kpoint%v(3), &
     &             bpath%bands(currkpoint)%energies(currBand)
          end select
        end do
      end do
      close (unit)
      end subroutine getwannierbands


!
! reportbrillouinpathspec_array --- write the specifications of a set of paths through the
!       Brillouin zone to fstdout
!

      subroutine reportbrillouinpathspec_array(bpaths, numBrillouinPaths, periodicity)
      implicit none

      integer, intent(in) :: numBrillouinPaths, periodicity
      type(tbrillouinpathspec), intent(in) :: bpaths(:)
      integer currPath

      write (*,'(A,I2,A)') ' Found ', numBrillouinPaths, &
     &          ' path(s) through the Brillouin zone'
      do currPath=1, numBrillouinPaths
        write (*,'(A,I2,A,I3,A,F5.3,A,F5.3,A)') ' Path No. ', currPath, &
     &          ' contains ', bpaths(currPath)%numkpoints,              &
     &          ' k-points and is ', bpaths(currPath)%metriclength,     &
     &          ' (metric), ', bpaths(currPath)%coordlength,            &
     &          ' (coordinates) long'
        call writebrillouinpathspec(bpaths(currPath), fstdout, periodicity)
        write (*,*)
      end do
      end subroutine reportbrillouinpathspec_array


!
! reportbrillouinpathspec_bs --- write the specifications of a set of paths through the
!       Brillouin zone to fstdout
!

      subroutine reportbrillouinpathspec_bs(bs)
      implicit none

      type(tbandstructure), intent(in) :: bs
      integer currPath

      write (*,'(A,I2,A)') ' Found ', bs%numPaths, &
     &          ' path(s) through the Brillouin zone'
      do currPath=1, bs%numPaths
        write (*,'(A,I2,A,I3,A,F5.3,A,F5.3,A)') ' Path No. ', currPath,    &
     &          ' contains ', bs%paths(currPath)%spec%numkpoints,          &
     &          ' k-points and is ', bs%paths(currPath)%spec%metriclength, &
     &          ' (metric), ', bs%paths(currPath)%spec%coordlength,        &
     &          ' (coordinates) long'
        call writebrillouinpathspec(bs%paths(currPath)%spec, fstdout, bs%periodicity)
        write (*,*)
      end do
      end subroutine reportbrillouinpathspec_bs


!
! fermienergy --- Determine the Fermi energy (top of valence band)
!       of solids with a band gap
!

      subroutine fermienergy(bs, fenergy, bandgap)
      implicit none

      type(tbandstructure), intent(in) :: bs
      real (dblprec), intent(out) :: fenergy, bandgap
      integer currPath, currkpoint

      fenergy = -huge(fenergy)
      bandgap =  huge(bandgap)

      if (bs%fock) then
        write (*,*) 'bs%numOccupied, bs%startOccupied'
        write (*,*) bs%numOccupied, bs%startOccupied

        if ((bs%numOccupied + bs%startOccupied - 1) >= bs%numOrbitals) then
          write (*,*) 'fermienergy: Cannot compute Fermi energy or band gap, no conduction band!'
          return
        end if

        do currPath=1, bs%numPaths
          do currkpoint=1, bs%paths(currPath)%spec%numkpoints
            ! Fermi energy: top of the highest valence band
            if (bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bs%numOccupied - bs%startOccupied + 1) > fenergy) then
              fenergy = bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bs%numOccupied - bs%startOccupied + 1)
            end if

            ! Band gap: bottom of the lowest conduction band
            if (bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bs%numOccupied + bs%startOccupied) < bandgap) then
              bandgap = bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bs%numOccupied + bs%startOccupied)
            end if
          end do
        end do

      else

        stop 'fermienergy: unsupported for correlated bands!'

/**********************************************
Fill in appropriate algorithm

        do currPath=1, bs%numPaths
          do currkpoint=1, bs%paths(currPath)%spec%numkpoints
            ! Search the band gap, bands are sorted occupied first
            do currBand=1, bs%paths(currPath)%bands(currkpoint)%numBands
              if (bs%paths(currPath)%bands(currkpoint)%energies(currBand) > 0) then
                if (bs%paths(currPath)%bands(currkpoint)%energies(bs%numOccupied) > fenergy) then
                  fenergy = bs%paths(currPath)%bands(currkpoint)%energies(max(1, currBand-1))
                end if
                exit
              end if
            end do
          end do
        end do
**********************************************/

      end if

      bandgap = bandgap - fenergy
      end subroutine fermienergy


!
! bandwidth --- Determine the band width of an energy band
!

      real (dblprec) function bandwidth(bs, bandno, bspmin, bspmax)
      implicit none

      type(tbandstructure), intent(in) :: bs
      integer, intent(in) :: bandno
      type(tbspoint), intent(out), optional :: bspmin, bspmax
      integer currPath, currkpoint
      real (dblprec) bandmin, bandmax

      bandmin =  huge(bandmin)
      bandmax = -huge(bandmax)

      if (bs%fock) then

        do currPath=1, bs%numPaths
          do currkpoint=1, bs%paths(currPath)%spec%numkpoints

            ! Minimum band energy: bottom of the band
            if (bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bandno - bs%startOccupied + 1) < bandmin) then
              bandmin = bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bandno - bs%startOccupied + 1)
              if (present(bspmin)) then
                 bspmin%pathno   = currPath
                 bspmin%kpointno = currkpoint
                 bspmin%bandno   = bandno
              end if
            end if

            ! Maximum band energy: top of the band
            if (bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bandno - bs%startOccupied + 1) > bandmax) then
              bandmax = bs%paths(currPath)%bands(currkpoint)%energies &
     &           (bandno - bs%startOccupied + 1)
              if (present(bspmax)) then
                 bspmax%pathno   = currPath
                 bspmax%kpointno = currkpoint
                 bspmax%bandno   = bandno
              end if
            end if

          end do
        end do

      else

        stop 'bandwidth: unsupported for correlated bands!'
      end if

      bandwidth = bandmax - bandmin
      end function bandwidth


!
! findkpoint --- search a given k-point in a path through the Brillouin zone
!

      integer function findkpoint(path,kp)
      implicit none

      type(tbrillouinpath), intent(in) :: path
      type(tlatticevector), intent(in) :: kp
      integer currkpoint
      type(tlatticevector) tkp

      ! Search the CO-ADC k-point in the set of k-points given
      ! in the Hartree-Fock band structure for cheating
      !write (*,*) 'Searching ', kp1, currBand + startquasiband, ' Path ', currPath
      !write (*,*) 'currBand = ', currBand, ', startquasiband = ', startquasiband
      do currkpoint = 1, path%spec%numkpoints
        tkp = path%bands(currkpoint)%kpoint



!        if (sqrt((kp1%a - kp2%a)**2 + (kp1%b - kp2%b)**2 &
!     &            + (kp1%c - kp2%c)**2) < 1.0D-6) then
!              if (sqrt((abs(kp1%v(1)) - abs(kp2%v(1)))**2 &
!     &                + (abs(kp1%v(2)) - abs(kp2%v(2)))**2 &
!     &               + (abs(kp1%v(3)) - abs(kp2%v(3)))**2) < 1.0D-6) then
!          if (sqrt((kp1%v(1) - kp2%v(1))**2 + (kp1%v(2) - kp2%v(2))**2 &
!     &        + (kp1%v(3) - kp2%v(3))**2) < 1.0D-6) then
!              write (*,*) 'Found ', kp1, 'Band ', currBand &
!     &                + startquasiband, ' Path ', currPath



        tkp%v = tkp%v - kp%v
        if (sqrt(dot_product(tkp%v,tkp%v)) < 1.0D-10) then
          findkpoint = currkpoint
          return
        end if
      end do

      ! k-point from the CO-ADC band structure not found in the
      ! Hartree-Fock band structure
      write (*,*) 'k-point ', kp, ' not found in path!'
      stop 'findkpoint: k-point not found in path!'
      end function findkpoint
      end module bandstructure