!
! startvec --- convert MOLPRO orbital description to a Wannier98 start vector
! Christian Buth
! 21.04.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
!
!
! Version 1.00: 26.09.2004 (Read in Wannier and MOLPRO parameters for a cluster
!       and write a start vector for the full cluster)
! Version 1.10: 03.10.2004 (Read in angular momentum of basis functions from
!       MOLPRO output, resort d-functions for Wannier)
! Version 1.20: 31.10.2004 (Correct resorting of D-functions, user interface)
! Version 1.30: 26.01.2006 (Fixed a bug in the determination of the number
!       of occupied orbitals in the molecular cluster)
!


      program startvec
      use fortrantl
      use physconst
      implicit none

      type geometry
        character(4) element
        integer cardinal, &     ! Cardinal number of the atom in the geometry output
     &          basisno,  &     ! Number of basis functions on the atom
     &          sgtono,   &     ! Number of basis functions of S-type on the atom
     &          pgtono,   &     ! Number of basis functions of P-type on the atom
     &          dgtono,   &     ! Number of basis functions of D-type on the atom
     &          cellno,   &     ! Number of the unit cell, the atom belongs to
     &          atomno          ! Atom number in the unit cell
        real(dblprec) nuccharge, x, y, z
      end type geometry

      type molecularorbital
        integer cellno
        real(dblprec) x, y, z
      end type molecularorbital

      character(1024) filemolpro,  &
     &                filewannier

      integer, parameter :: umolpro=101
      integer i, currMOOcc, numAtoms, numBasis, numMOOcc, numWAOcc, iatom
      integer currWAOcc, numUnitCells, numLattices
      integer wanniermo, idx, imp, iuc, ibas, ilat, ibasoff, numBasisCell
      integer iminatom, idummy
      integer, allocatable :: selectMO(:), sortMO(:), basismolpro(:), basisoffsetmolpro(:)

      real(dblprec) orbcard, occupation, orbenergy, dummy, dist, mindist
      real(dblprec), allocatable :: orbcoeff(:,:), wastart(:)
      type(geometry), allocatable :: molpro(:), wannier(:)
      type(molecularorbital), allocatable :: orbitals(:)

      namelist /start/ filemolpro, filewannier


      write (*,*) 'Startvector generator for Wannier98 using MOLPRO'
      write (*,*) 'V1.30 (c) 2004 Christian Buth'
      write (*,*)

      read(*,start)

      call readwannier ! (wannier)
      call readmolpro  ! (molpro)

      ! Resort MOLPRO atoms
      allocate(sortMO(numAtoms))
      sortMO = 0
      do iatom=1, numAtoms
        searchatom: do i=1, numAtoms
          dist =        (molpro(i)%x - wannier(iatom)%x)**2
          dist = dist + (molpro(i)%y - wannier(iatom)%y)**2
          dist = dist + (molpro(i)%z - wannier(iatom)%z)**2
          dist = sqrt(dist)
          if (dist < 1.0E-5) then
            sortMO(iatom) = i
            wannier(iatom)%basisno = molpro(i)%basisno
            wannier(iatom)%sgtono  = molpro(i)%sgtono
            wannier(iatom)%pgtono  = molpro(i)%pgtono
            wannier(iatom)%dgtono  = molpro(i)%dgtono
            molpro(i)%atomno = wannier(iatom)%atomno
            molpro(i)%cellno = wannier(iatom)%cellno
            exit searchatom
          end if
        end do searchatom
        if (sortMO(iatom) == 0) then
          stop 'Atom not found in MOLPRO geometry!'
        end if
        !write (*,*) 'Wannier atom ', iatom, ' is MOLPRO atom ', sortMO(iatom)
      end do

      ! Print numbers of basis functions
      do iatom=1, numAtoms
!        write (*,*) 'Number of basis functions on atom ', iatom, ' is ', &
!     &               wannier(iatom)%basisno, ' S-type ', wannier(iatom)%sgtono, &
!     &               ' P-type ', wannier(iatom)%pgtono, ' D-type ', wannier(iatom)%dgtono
      end do
      do ilat=1, numLattices
        iatom = 1 + (ilat-1) * numUnitCells
        write (*,*) 'Number of basis functions on atom ', wannier(iatom)%basisno, &
     &              ' S-type ', wannier(iatom)%sgtono, ' P-type ', wannier(iatom)%pgtono, &
     &              ' D-type ', wannier(iatom)%dgtono, ' in lattice ', ilat
      end do
      numBasisCell = numBasis / numUnitCells
      write (*,*) 'Number of basis functions per unit cell ', numBasisCell


      ! Associate localized molecular orbital of MOLPRO with unit cells
      do currMOOcc=1, numMOOcc
        mindist = huge(mindist)
        iminatom = 0
        do iatom=1, numAtoms
          dist =        (molpro(iatom)%x - orbitals(currMOOcc)%x)**2
          dist = dist + (molpro(iatom)%y - orbitals(currMOOcc)%y)**2
          dist = dist + (molpro(iatom)%z - orbitals(currMOOcc)%z)**2
          dist = sqrt(dist)
          if (dist < mindist) then
            mindist = dist
            iminatom = iatom
          end if
        end do
        orbitals(currMOOcc)%cellno = molpro(iminatom)%cellno
        !write (*,*) 'Wannier atom ', iatom, ' is MOLPRO atom ', sortMO(iatom)
      end do

      do iuc=1, numUnitCells
        write (*,*) 'Localized MOs in unit cell ', iuc,':'
        do currMOOcc=1, numMOOcc
          if (orbitals(currMOOcc)%cellno == iuc) then
            write (*,'(I4,A)',advance='no') currMOOcc, ','
          end if
        end do
        write (*,*)
      end do

      allocate(selectMO(numMOOcc))
      currWAOcc = 1
      do currMOOcc=1, numMOOcc
        if (orbitals(currMOOcc)%cellno == 1) then
          selectMO(currWAOcc) = currMOOcc
          currWAOcc = currWAOcc + 1
        end if
      end do

      write (*,'(A)', advance='no') 'Selected localized MOs in origin cell '
      do currWAOcc=1, numWAOcc
        write (*,'(I4,A)',advance='no') selectMO(currWAOcc), ','
      end do
      write (*,*)

      allocate(wastart(numBasis))

      ! Write header of orbital coefficients file
      open(unit=umolpro, file='ORB.OUT', action='write', status='replace')
      idummy = 0
      write (umolpro,*) numWAOcc, numBasis, idummy
      dummy = 2
      write (umolpro,*) (dummy, i=1, numWAOcc)


      ! Determine the mapping between basis function in Wannier
      ! and basis function in MOLPRO
      allocate(basisoffsetmolpro(numAtoms))
      ibasoff = 0
      do iatom=1, numAtoms
        basisoffsetmolpro(iatom) = ibasoff
        !write (*,*) 'iatom ', iatom, ' basisoffsetmolpro(iatom)', ibasoff
        if (iatom <= numAtoms) then
          ibasoff = ibasoff + molpro(iatom)%basisno
        end if
      end do

      ! Resort basis functions of MOLPRO to match the sorting in Wannier
      allocate(basismolpro(numBasis))
      ibas = 1
      do iatom=1, numAtoms
        do i=1, molpro(iatom)%sgtono
          basismolpro(ibas) = ibas
          ! Invalidate
          !if (molpro(iatom)%cellno /= 1) then
          !  basismolpro(ibas) = 0
          !end if
          ibas = ibas + 1
        end do

        do i=1, molpro(iatom)%pgtono
          basismolpro(ibas) = ibas
          ! Invalidate
          !if (molpro(iatom)%cellno /= 1) then
          !  basismolpro(ibas) = 0
          !end if
          ibas = ibas + 1
        end do

        ! There are 5 d-functions per atomic shell
        ! (use "spherical" keyword in MOLPRO)
        do i=1, molpro(iatom)%dgtono / 5
          ! Invalidate
          !if (molpro(iatom)%cellno /= 1) then
          !  basismolpro(ibas : ibas + 4) = 0
          !end if

          ! Mein Schema (Hilfe von Hermann Stoll, Martin Albrecht)
          ! MOLPRO and Wannier98 use solid harmonics
          ! MOLPRO orbital coefficients: 3d0    3d2-    3d1+    3d2+    3d1-
          ! # / Wannier98 Koeffizient / MOLPRO Koeffizient / #
          ! 1. d_x^2-y^2 = d_(+2)  4.
          ! 2. d_xy      = d_(-2)  2.
          ! 3. d_xz      = d_(+1)  3.
          ! 4. d_yz      = d_(-1)  5.
          ! 5. d_z^2     = d_( 0)  1.
          basismolpro(ibas + 0) = ibas + 3
          basismolpro(ibas + 1) = ibas + 1
          basismolpro(ibas + 2) = ibas + 2
          basismolpro(ibas + 3) = ibas + 4
          basismolpro(ibas + 4) = ibas + 0
          ibas = ibas + 5
        end do
      end do
      write (*,*) 'Resorted ', ibas-1, ' numBasis ', numBasis

      ! Generate and write Wannier orbitals for Wannier
      do currWAOcc=1, numWAOcc
        wanniermo = selectMO(currWAOcc)
        idx = 1

        do ilat=1, numLattices
          do iuc=1, numUnitCells
            do ibas=1, wannier((ilat-1) * numUnitCells + 1)%basisno
              imp = numUnitCells * (ilat-1) + iuc       ! Determine atom in the cluster of Wannier
              imp = sortMO(imp)                         ! Determine atom in the cluster of MOLPRO
              imp = basisoffsetmolpro(imp) + ibas       ! Determine the position of the first basis
                                                        ! function of the atom in MOLPRO
              imp = basismolpro(imp)                    ! Determine basis function MOLPRO (resorting)
              if (imp /= 0) then
                wastart(idx) = orbcoeff(imp, wanniermo)
              else
                wastart(idx) = 0
              end if
              idx = idx + 1
            end do
          end do
        end do

        write (umolpro,*) (wastart(i), i=1, numBasis)
      end do

      deallocate(selectMO, sortMO, basismolpro, basisoffsetmolpro, &
     &           orbcoeff, wastart, molpro, wannier, orbitals)


!
! Internal procedures
!

      contains


!
! readwannier --- read geometry of Wannier cluster from output
!

      subroutine readwannier !(wannier)
      implicit none

      !type(geometry), intent(out) :: wannier(:)

      character(132) line
      character(*), parameter :: symopstr =    ' no. of symmetry operations.:', &
     &                           nlatstr =     ' sr basgen: nlat was read to',  &
     &                           occorbstr =   ' A Closed-Shell Hartree-Fock',  &
     &                           unitcellstr = ' NO. OF UNIT CELLS WHOSE BASI'
      integer iatom, atomno, ilat, iuc
      integer, parameter :: uwannier=100
      real(dblprec) nucmass

      open(unit=uwannier, file=filewannier, action='read', status='old')

      do
        read (uwannier,'(A29)') line
        if (trim(line) == symopstr) then
          read (uwannier,'(A15,I)') line, numAtoms
          read (uwannier,*)
          read (uwannier,*)
          read (uwannier,*)
          write (*,*) 'Geometry found in Wannier output '
          write (*,*)
          write (*,*) 'Number of atoms', numAtoms
          write (*,*) 'Coordinates of atoms in the Wannier98 cluster'
          allocate(wannier(numAtoms))

          do iatom=1, numAtoms
            read (uwannier,'(I4,A4,5F15.8)') wannier(iatom)%cardinal,       &
     &          wannier(iatom)%element, wannier(iatom)%x, wannier(iatom)%y, &
     &          wannier(iatom)%z, wannier(iatom)%nuccharge, nucmass
!            write (*,'(A2,3(A,ES15.8))') adjustl(wannier(iatom)%element), &
!     &          ' ,', wannier(iatom)%x * BohrtoAngstroem,                 &
!     &          ' ,', wannier(iatom)%y * BohrtoAngstroem,                 &
!     &          ' ,', wannier(iatom)%z * BohrtoAngstroem
          end do
        end if

        if (trim(line) == nlatstr) then
          read (uwannier,'(A23,I)') line, numLattices
          write (*,*) 'Number of atoms per unit cell ', numLattices
        end if

        if (trim(line) == unitcellstr) then
          read (uwannier,'(A38,I)') line, numUnitCells
          write (*,*) 'Number of unit cells ', numUnitCells
        end if

        if (trim(line) == occorbstr) then
          read (uwannier,'(A29,I)') line, numWAOcc
          write (*,*) 'Number of occupied Wannier orbitals ', numWAOcc
          close(uwannier)

          ! Set the number of the atom in the unit cell
          atomno = 1
          iatom  = 1
          do ilat=1, numLattices
            do iuc=1, numUnitCells
              !write (*,*) 'atom ', iatom, ' atomno ', atomno
              wannier(iatom)%atomno = atomno
              wannier(iatom)%cellno = iuc
              iatom = iatom + 1
            end do
            atomno = atomno + 1
          end do
          return
        end if
      end do

      stop 'Error: not a Wannier98 output'
      end subroutine readwannier


!
! readmolpro --- read data from MOLPRO output
!

      subroutine readmolpro !(molpro)
      implicit none

      ! type(geometry), intent(out) :: molpro(:)
      character gtotype, prevgtotype
      character(132) line
      character(*), parameter :: sewardstr   = ' ATOMIC COORDINATES', &
     &                           nuclchrgstr = ' NUCLEAR CHARGE:',    &
     &                           orbsstr1    = ' Localized orbitals', &
     &                           orbsstr2    = ' ------------------', &
     &                           centersstr  = ' Centers of charge',  &
     &                           basisbegstr = ' BASIS DATA',         &
     &                           basisendstr = ' NUCLEAR CHARGE:'
      integer iatom, prevatom, symmetryno, orbitalno, basisno, shellno
      integer, parameter :: umolpro=102
      real basid

      open(unit=umolpro, file=filemolpro, action='read', status='old')
      allocate(molpro(numAtoms))

      do
        read (umolpro,'(A19)') line

        ! Count number of basis functions per angular momentum and atom
        ! For this to work 'gprint,basis' has to be specified in the
        ! MOLPRO input
        if (trim(line) == basisbegstr) then
          read (umolpro,*)
          read (umolpro,*)
          read (umolpro,*)
          write (*,*) 'Basis set found in MOLPRO output '

          do iatom=1, numAtoms
            molpro(iatom)%basisno = 0
            molpro(iatom)%sgtono  = 0
            molpro(iatom)%pgtono  = 0
            molpro(iatom)%dgtono  = 0
          end do
          basisno = 1

          countbasis: do
            read (umolpro,'(A19)')
            read (umolpro,'(A19)') line
            if (trim(line) == basisendstr) then
              exit countbasis
            else
              backspace(umolpro)
              backspace(umolpro)
            end if

            read (umolpro,*) basid
            if (abs(real(basisno) + 0.1 - basid) < 1.0E-6) then
              backspace(umolpro)
              prevatom    = iatom
              prevgtotype = gtotype
              read (umolpro,'(F6.1,A3,I5,I3,A1)') basid, line, iatom, shellno, gtotype
              if (iatom == 0) then
                iatom   = prevatom
                gtotype = prevgtotype
              end if
              ! write (*,*) 'Basis function ', basid, trim(line), iatom, shellno, gtotype
              molpro(iatom)%basisno = molpro(iatom)%basisno + 1
              select case (gtotype)
                case ('s')
                  molpro(iatom)%sgtono = molpro(iatom)%sgtono + 1
                case ('p')
                  molpro(iatom)%pgtono = molpro(iatom)%pgtono + 1
                case ('d')
                  molpro(iatom)%dgtono = molpro(iatom)%dgtono + 1
                case default
                  stop 'Unknown GTO angular momentum'
              end select
              basisno = basisno + 1
            end if
          end do countbasis
        end if

        ! Read MOLPRO cluster geometry
        if (trim(line) == sewardstr) then
          read (umolpro,*)
          read (umolpro,*)
          read (umolpro,*)
          write (*,*) 'Geometry found in MOLPRO output '

          do iatom=1, numAtoms
            read (umolpro,'(I4,A4,F11.2,3F15.9)') molpro(iatom)%cardinal, &
     &          molpro(iatom)%element, molpro(iatom)%nuccharge,     &
     &          molpro(iatom)%x, molpro(iatom)%y, molpro(iatom)%z
!            write (*,'(A2,3(A,ES15.8))') adjustl(molpro(iatom)%element), &
!     &          ' ,', molpro(iatom)%x * BohrtoAngstroem,                 &
!     &          ' ,', molpro(iatom)%y * BohrtoAngstroem,                 &
!     &          ' ,', molpro(iatom)%z * BohrtoAngstroem
          end do
        end if

        if (trim(line) == nuclchrgstr) then
          backspace(umolpro)
          read (umolpro,'(A29,I8)') line, numMOOcc
          numMOOcc = numMOOcc / 2
          write (*,*) 'Number of occupied orbitals in cluster ', numMOOcc
          read (umolpro,*)
          read (umolpro,*)
          read (umolpro,'(A25,I)') line, numBasis
          write (*,*) 'Number of basis functions ', numBasis
        end if

        ! Read orbital coefficients
        if (trim(line) == orbsstr1) then
          read (umolpro,'(A19)') line
          if (trim(line) == orbsstr2) then
            allocate(orbcoeff(numBasis, numMOOcc))
            write (*,*) 'Found orbital coefficients of localized orbitals'

            ! Skip empty lines and description of coefficients
            do i=1, 4 + numBasis / 10
              read (umolpro,*)
            end do

            do currMOOcc=1, numMOOcc
              read (umolpro,*) orbcard, occupation, orbenergy, &
     &                         (orbcoeff(i, currMOOcc), i=1, numBasis)
              !write (*,*) orbcard, occupation, orbenergy
            end do
          end if
        end if

        ! Read centers of charge of the localized orbitals
        if (trim(line) == centersstr) then
          write (*,*) 'Found centers of charge of the localized orbitals'
          read (umolpro,*)
          read (umolpro,*)
          allocate(orbitals(numMOOcc))
          do currMOOcc=1, numMOOcc
            read (umolpro,*) symmetryno, orbitalno, orbitals(currMOOcc)%x, &
     &                       orbitals(currMOOcc)%y, orbitals(currMOOcc)%z
!            write (*,*) symmetryno, orbitalno, orbitals(currMOOcc)%x, &
!     &                  orbitals(currMOOcc)%y, orbitals(currMOOcc)%z
          end do
          close(umolpro)
          return
        end if
      end do

      stop 'Error: not a MOLPRO output'
      end subroutine readmolpro
      end program startvec
